Type: Package
Package: later2
Title: Utilities for Scheduling Functions to Execute Later with Event
        Loops
Version: 0.1
Authors@R: c(
    person("Winston", "Chang", , "winston@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-1576-2126")),
    person("Joe", "Cheng", , "joe@posit.co", role = "aut"),
    person("Charlie", "Gao", , "charlie.gao@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-0750-061X")),
    person("Posit Software, PBC", role = c("aut", "cph"),
           comment = c(ROR = "03wc8by49")),
    person("Marcus", "Geelnard", role = c("ctb", "cph"),
           comment = "TinyCThread library, https://tinycthread.github.io/"),
    person("Evan", "Nemerson", role = c("ctb", "cph"),
           comment = "TinyCThread library, https://tinycthread.github.io/"),
    person(
        given = "Mauricio", family = "Vargas Sepulveda", role = c("aut", "cre"),
        email = "m.vargas.sepulveda@gmail.com",
        comment = c(ORCID = "0000-0003-1017-7574"))
  )
Description: Executes arbitrary R or C functions some time after the current
    time, after the R execution stack has emptied. The functions are scheduled
    in an event loop. This is a derived work from the 'later' package aiming
    to reduce the number of dependencies.
License: Apache License 2.0
URL: https://github.com/pachadotdev/later2
BugReports: https://github.com/pachadotdev/later2/issues
Depends: R (>= 3.5)
Suggests: tinytest, parallel, litedown
LinkingTo: cpp4r
Encoding: UTF-8
VignetteBuilder: litedown
NeedsCompilation: yes
Packaged: 2026-08-23 18:00:28 UTC; pacha
Author: Winston Chang [aut] (ORCID: <https://orcid.org/0000-0002-1576-2126>),
  Joe Cheng [aut],
  Charlie Gao [aut] (ORCID: <https://orcid.org/0000-0002-0750-061X>),
  Posit Software, PBC [aut, cph] (ROR: <https://ror.org/03wc8by49>),
  Marcus Geelnard [ctb, cph] (TinyCThread library,
    https://tinycthread.github.io/),
  Evan Nemerson [ctb, cph] (TinyCThread library,
    https://tinycthread.github.io/),
  Mauricio Vargas Sepulveda [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1017-7574>)
Maintainer: Mauricio Vargas Sepulveda <m.vargas.sepulveda@gmail.com>
Repository: CRAN
Date/Publication: 2026-09-03 12:50:02 UTC

# Generated by tinyroxygen: do not edit by hand
useDynLib(later2, .registration=TRUE)
S3method(format, event_loop)
S3method(format, promise)
S3method(print, event_loop)
S3method(print, promise)
export(catch)
export(create_loop)
export(current_loop)
export(destroy_loop)
export(exists_loop)
export(finally)
export(global_loop)
export(is.promise)
export(later)
export(later_fd)
export(loop_empty)
export(next_op_secs)
export(promise)
export(run_now)
export(then)
export(with_loop)
export(with_temp_loop)

# later2

<!-- badges: start -->

<!-- badges: end -->

Schedule an R function or formula to run after a specified period of time. This is similar to JavaScript's `setTimeout` function.
Like JavaScript, R is single-threaded so there's no guarantee that the operation will run exactly at the requested time, only that
at least that much time will elapse.

To avoid bugs due to reentrancy, by default, scheduled operations only run when there is no other R code present on the execution
stack; i.e., when R is sitting at the top-level prompt. You can force past-due operations to run at a time of your choosing by
calling `later2::run_now()`.

This package is derived from [later](https://github.com/r-lib/later), which uses derived work from
[background](https://github.com/s-u/background) package and similar code in
[Rhttpd](https://search.r-project.org/CRAN/refmans/Rook/html/Rhttpd-class.html).

## Installation

You can install the development version of later with:

``` r
pak::pak("pachadotdev/later2")
```

The package is not currently on CRAN.

## Usage from R

Pass a function (in this case, delayed by 5 seconds):

``` r
later2::later(\() print("Got here!"), 5)
```

Or a formula (in this case, run as soon as control returns to the top-level):

``` r
later2::later(~print("Got here!"))
```

### File Descriptor Readiness

It is also possible to have a function run based on when file descriptors are ready for reading or
writing, at an unknown time in the future.

Below, a logical vector is printed indicating which of file descriptors 21 or 22 were ready,
subject to a timeout of 1s.

``` r
later2::later_fd(\(x) print(x), c(21L, 22L), timeout = 1)
```

This is useful in particular for asynchronous I/O, allowing reads to be made from TCP sockets as
soon as data becomes available. Functions such as `curl::multi_fdset()` return the relevant file
descriptors to be monitored.

## Usage from C++

You can also call `later2::later` from C++ code in your own packages, to cause your own C-style
functions to be called back. This is safe to call from either the main R thread or a different
thread; in both cases, your callback will be invoked from the main R thread.

`later2::later` is accessible from `later2_api.h` and its prototype looks like this:

``` cpp
void later(void (*func)(void*), void* data, double secs)
```

The first argument is a pointer to a function that takes one `void*` argument and returns void. The
second argument is a `void*` that will be passed to the function when it's called back. And the
third argument is the number of seconds to wait (at a minimum) before invoking.

`later2::later_fd` is also accessible from `later2_api.h` and its prototype looks like this:

``` cpp
void later_fd(void (*func)(int *, void *), void *data, int num_fds, struct pollfd *fds, double secs)
```

The first argument is a pointer to a function that takes two arguments:

1. An `int*` array provided by `later_fd()` when called back, and the second being
  a `void*`. The `int*` array will be the length of `num_fds` and contain the values `0`, `1` or
  `NA_INTEGER` to indicate the readiness of each file descriptor, or an error condition
  respectively.
2. A `data` object passed to the `void*` argument of the function when it's called back. The other
  required arguments are the total number of file descriptors, a pointer to an array of
  `stuct pollfd`, and the number of seconds to wait until timing out.

To use the C++ interface, you'll need to add `later` to your `DESCRIPTION` file under
`Imports` and `LinkingTo`, and also make sure that your `NAMESPACE` file has an `import(later)`
entry. See the [later2test](https://github.com/pachadotdev/later2/tree/main/later2test) side package
which I created to test later2 to check how to use it from another package.

### Background tasks

Finally, this package also offers a higher-level C++ helper class to make it easier to execute tasks
on a background thread. It is also available from `later2_api.h` and its public/protected interface
looks like this:

``` cpp
class BackgroundTask {

public:
  BackgroundTask();
  virtual ~BackgroundTask();

  // Start executing the task
  void begin();

protected:
  // The task to be executed on the background thread.
  // Neither the R runtime nor any R data structures may be
  // touched from the background thread; any values that need
  // to be passed into or out of the Execute method must be
  // included as fields on the Task subclass object.
  virtual void execute() = 0;

  // A short task that runs on the main R thread after the
  // background task has completed. It's safe to access the
  // R runtime and R data structures from here.
  virtual void complete() = 0;
}
```

Create your own subclass, implementing a custom constructor plus the `execute` and `complete`
methods.

It's critical that the code in your `execute` method not mutate any R data structures, call any R
code, or cause any R allocations, as it will execute in a background thread where such operations
are unsafe. You can, however, perform such operations in the constructor (assuming you perform
construction only from the main R thread) and `complete` method. Pass values between the constructor
and methods using fields.

``` cpp
#include "cpp11.hpp"
#include <later2_api.h>

using namespace cpp11;

class MyTask : public later2::BackgroundTask {
public:
  MyTask(doubles vec) :
    inputVals(as_cpp<std::vector<double>>(vec)) {
  }

protected:
  void execute() {
    double sum = 0;
    for (std::vector<double>::const_iterator it = inputVals.begin();
      it != inputVals.end();
      it++) {

      sum += *it;
    }
    result = sum / inputVals.size();
  }

  void complete() {
    Rprintf("Result is %f\n", result);
  }

private:
  std::vector<double> inputVals;
  double result;
};
```

To run the task, `new` up your subclass and call `begin()`, for example
`(new MyTask(vec))->begin()`. There's no need to keep track of the pointer; the task object will
delete itself when the task is complete.

``` r
[[cpp11::register]] void asyncMean(doubles data) {
  (new MyTask(data))->begin();
}
```

## Example with parallel jobs (Unix-like systems)

Using the base R 'parallel' package, you can fork child processes and obtain real, pollable file
descriptors. A child's `fd` becomes ready for reading once the child finishes running.

```r
# 1. timeout: prints FALSE, FALSE ----
job1 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
job2 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
fd1 <- job1$fd[1]
fd2 <- job2$fd[1]
later_fd(print, c(fd1, fd2), timeout = 0.1)
Sys.sleep(0.2)
run_now()

# 2. fd1 ready: prints TRUE, FALSE ----
job1 <- parallel::mcparallel(TRUE)
fd1 <- job1$fd[1]
Sys.sleep(0.1)
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 3. both ready: prints TRUE, TRUE ----
job2 <- parallel::mcparallel(TRUE)
fd2 <- job2$fd[1]
Sys.sleep(0.1)
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 4. fd2 ready: prints FALSE, TRUE ----
parallel::mccollect(job1)
job1 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
fd1 <- job1$fd[1]
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 5. fds invalid: prints NA, NA ----
parallel::mccollect(job1)
parallel::mccollect(job2)
later_fd(print, c(fd1, fd2), timeout = 0)
Sys.sleep(0.1)
run_now()
```

## Example with "promise" abstractions

I adapted some functions from the 'promises' package using base R. As 'later' points out, it is
useful to execute tasks on background threads if you cannoy get the results back in R.
See the [later2test](https://github.com/pachadotdev/later2/tree/main/later2test) for more "promise"
abstractions such as:

```r
# test that resolve works
result <- 0
later2::promise(function(resolve, reject) {
  later2test::asyncFib(resolve, reject, 3)
}) |>
  later2::then(\(x) {
    result <<- x
  })

expect_identical(result, 0)
later2::run_now(1)
while (!later2::loop_empty()) {
  later2::run_now(0.1)
}
expect_identical(result, 2)
```

#' @title Utilities for Scheduling Functions to Execute Later with Event Loops
#' @description Executes arbitrary R or C functions some time after the current
#'  time, after the R execution stack has emptied. The functions are scheduled
#'  in an event loop. This is a derived work from the 'later' package aiming
#'  to reduce the number of dependencies.
#' @useDynLib later2, .registration=TRUE
"_PACKAGE"

# Generated by cpp4r: do not edit by hand

testCallbackOrdering <- function() {
	invisible(.Call(`_later2_testCallbackOrdering`))
}

log_level <- function(level) {
	.Call(`_later2_log_level`, level)
}

using_ubsan <- function() {
	.Call(`_later2_using_ubsan`)
}

execLater_fd <- function(callback, readfds, writefds, exceptfds, timeoutSecs, loop_id) {
	.Call(`_later2_execLater_fd`, callback, readfds, writefds, exceptfds, timeoutSecs, loop_id)
}

fd_cancel <- function(xptr) {
	.Call(`_later2_fd_cancel`, xptr)
}

setCurrentRegistryId <- function(id) {
	invisible(.Call(`_later2_setCurrentRegistryId`, id))
}

getCurrentRegistryId <- function() {
	.Call(`_later2_getCurrentRegistryId`)
}

deleteCallbackRegistry <- function(loop_id) {
	.Call(`_later2_deleteCallbackRegistry`, loop_id)
}

notifyRRefDeleted <- function(loop_id) {
	.Call(`_later2_notifyRRefDeleted`, loop_id)
}

createCallbackRegistry <- function(id, parent_id) {
	invisible(.Call(`_later2_createCallbackRegistry`, id, parent_id))
}

existsCallbackRegistry <- function(id) {
	.Call(`_later2_existsCallbackRegistry`, id)
}

list_queue_ <- function(id) {
	.Call(`_later2_list_queue_`, id)
}

execCallbacks <- function(timeoutSecs, runAll, loop_id) {
	.Call(`_later2_execCallbacks`, timeoutSecs, runAll, loop_id)
}

idle <- function(loop_id) {
	.Call(`_later2_idle`, loop_id)
}

ensureInitialized <- function() {
	invisible(.Call(`_later2_ensureInitialized`))
}

execLater <- function(callback, delaySecs, loop_id) {
	.Call(`_later2_execLater`, callback, delaySecs, loop_id)
}

cancel <- function(callback_id_s, loop_id) {
	.Call(`_later2_cancel`, callback_id_s, loop_id)
}

nextOpSecs <- function(loop_id) {
	.Call(`_later2_nextOpSecs`, loop_id)
}

new_weakref <- function(x) {
	.Call(`_later2_new_weakref`, x)
}

wref_key <- function(x) {
	.Call(`_later2_wref_key`, x)
}

# A minimal, dependency-free re-implementation of the core `promises`
# package API (`promise()`, `then()`, `catch()`, `finally()`, `is.promise()`),
# built entirely on later2's own event loop.
#
# The `promises` package schedules its internal fulfill/reject callbacks via
# `later::later()`, which runs on the `later` package's own C event loop.
# When code is driven by later2 (e.g. via `later2::run_now()`), those
# callbacks are registered on a completely different event loop and never
# fire. This file provides a self-contained implementation that calls
# later2's own `later()`, so promise resolution is properly synchronized
# with `run_now()`/`loop_empty()`.
#
# Unlike `promises`, this implementation has no dependencies (no R6,
# fastmap, rlang, magrittr, lifecycle) and does not support promise domains
# or `future`/`mirai` integration.

Promise <- function() {
  self <- new.env(parent = emptyenv())
  private <- new.env(parent = emptyenv())

  private$state <- "pending"
  private$value <- NULL
  private$visible <- TRUE
  private$publicResolveRejectCalled <- FALSE
  private$onFulfilled <- list()
  private$onRejected <- list()
  private$rejectionHandled <- FALSE

  private$doResolve <- function(value) {
    val <- withVisible(value)
    value <- val$value
    visible <- val$visible

    if (is.promise(value)) {
      impl <- attr(value, "promise_impl", exact = TRUE)
      if (identical(self, impl)) {
        return(private$doReject(simpleError(
          "Chaining cycle detected for promise"
        )))
      }
      value$then(private$doResolve, private$doReject)
    } else {
      private$doResolveFinalValue(value, visible)
    }
  }

  private$doReject <- function(reason) {
    if (is.promise(reason)) {
      reason$then(private$doResolve, private$doReject)
    } else {
      private$doRejectFinalReason(reason)
    }
  }

  # These "final" versions of resolve/reject are for when we've established
  # that the value/reason is not itself a promise.
  private$doResolveFinalValue <- function(value, visible) {
    private$value <- value
    private$visible <- visible
    private$state <- "fulfilled"

    later(function() {
      lapply(private$onFulfilled, function(f) {
        f(private$value, private$visible)
      })
      private$onFulfilled <- list()
    })
  }

  private$doRejectFinalReason <- function(reason) {
    private$value <- reason
    private$state <- "rejected"

    later(function() {
      lapply(private$onRejected, function(f) {
        private$rejectionHandled <- TRUE
        f(private$value)
      })
      private$onRejected <- list()

      later(function() {
        if (!private$rejectionHandled) {
          cat(
            file = stderr(),
            "Unhandled promise error: ",
            conditionMessage(reason),
            "\n",
            sep = ""
          )
        }
      })
    })
  }

  self$status <- function() {
    private$state
  }

  self$resolve <- function(value) {
    # Only allow this to be called once, then no-op.
    if (private$publicResolveRejectCalled) {
      return(invisible())
    }
    private$publicResolveRejectCalled <- TRUE

    tryCatch(
      {
        # Important: Do not trigger evaluation of value before passing to
        # doResolve. doResolve calls withVisible() on value, so evaluating
        # it before that point would cause the visibility to be lost.
        private$doResolve(value)
      },
      error = function(err) {
        private$doReject(err)
      }
    )

    invisible()
  }

  self$reject <- function(reason) {
    # Only allow this to be called once, then no-op.
    if (private$publicResolveRejectCalled) {
      return(invisible())
    }
    private$publicResolveRejectCalled <- TRUE

    tryCatch(
      {
        force(reason)
        if (is.character(reason)) {
          reason <- simpleError(reason)
        }
        private$doReject(reason)
      },
      error = function(err) {
        private$doReject(err)
      }
    )

    invisible()
  }

  self$then <- function(onFulfilled = NULL, onRejected = NULL, onFinally = NULL) {
    onFulfilled <- normalizeOnFulfilled(onFulfilled)
    onRejected <- normalizeOnRejected(onRejected)
    if (!is.function(onFinally)) {
      onFinally <- NULL
    }

    if (!is.null(onFinally)) {
      if (!is.null(onFulfilled) || !is.null(onRejected)) {
        stop(
          "A single `then` call cannot combine `onFinally` with `onFulfilled`/`onRejected`"
        )
      }
      spliced <- spliceOnFinally(onFinally)
      onFulfilled <- spliced$onFulfilled
      onRejected <- spliced$onRejected
    }

    promise2 <- promise(function(resolve, reject) {
      handleFulfill <- function(value, visible) {
        if (is.function(onFulfilled)) {
          resolve(onFulfilled(value, visible))
        } else {
          resolve(if (visible) value else invisible(value))
        }
      }

      handleReject <- function(reason) {
        if (is.function(onRejected)) {
          # Yes, resolve, not reject.
          resolve(onRejected(reason))
        } else {
          # Yes, reject, not resolve.
          reject(reason)
        }
      }

      if (private$state == "pending") {
        private$onFulfilled <- c(private$onFulfilled, list(handleFulfill))
        private$onRejected <- c(private$onRejected, list(handleReject))
      } else if (private$state == "fulfilled") {
        later(function() {
          handleFulfill(private$value, private$visible)
        })
      } else if (private$state == "rejected") {
        later(function() {
          private$rejectionHandled <- TRUE
          handleReject(private$value)
        })
      } else {
        stop("Unexpected state ", private$state)
      }
    })

    invisible(promise2)
  }

  self$catch <- function(onRejected) {
    invisible(self$then(onRejected = onRejected))
  }

  self$finally <- function(onFinally) {
    invisible(self$then(onFinally = onFinally))
  }

  self$format <- function() {
    if (private$state == "pending") {
      "<Promise [pending]>"
    } else {
      classname <- class(private$value)[[1]]
      if (length(classname) == 0) {
        classname <- ""
      }
      sprintf("<Promise [%s: %s]>", private$state, classname)
    }
  }

  class(self) <- "Promise"
  self
}

normalizeOnFulfilled <- function(onFulfilled) {
  if (!is.function(onFulfilled)) {
    if (!is.null(onFulfilled)) {
      warning("`onFulfilled` must be a function or `NULL`")
    }
    return(NULL)
  }

  args <- formals(onFulfilled)
  arg_count <- length(args)

  if (arg_count >= 2 && names(args)[[2]] == ".visible") {
    onFulfilled
  } else if (arg_count > 0) {
    function(value, .visible) {
      if (isTRUE(.visible)) {
        onFulfilled(value)
      } else {
        onFulfilled(invisible(value))
      }
    }
  } else {
    function(value, .visible) {
      onFulfilled()
    }
  }
}

normalizeOnRejected <- function(onRejected) {
  if (!is.function(onRejected)) {
    if (!is.null(onRejected)) {
      warning("`onRejected` must be a function or `NULL`")
    }
    return(NULL)
  }

  args <- formals(onRejected)
  arg_count <- length(args)

  if (arg_count >= 1) {
    onRejected
  } else {
    function(reason) {
      onRejected()
    }
  }
}

spliceOnFinally <- function(onFinally) {
  list(
    onFulfilled = finallyToFulfilled(onFinally),
    onRejected = finallyToRejected(onFinally)
  )
}

finallyToFulfilled <- function(onFinally) {
  force(onFinally)
  function(value, .visible) {
    onFinally()
    if (.visible) value else invisible(value)
  }
}

finallyToRejected <- function(onFinally) {
  force(onFinally)
  function(reason) {
    onFinally()
    stop(reason)
  }
}

#' @title Create a new promise object
#'
#' @description `promise()` creates a new promise, synchronized with later2's
#' event loop (instead of the `later` package's event loop, as the `promises`
#' package does). A promise is a placeholder object for the eventual result (or
#' error) of an asynchronous operation.
#'
#' The `action` function should be a piece of code that returns quickly, but
#' initiates a potentially long-running, asynchronous task. If/when the task
#' successfully completes, call `resolve(value)` where `value` is the result of
#' the computation. If the task fails, call `reject(reason)`, where `reason` is
#' either an error object or a character string.
#'
#' @param action A function with signature `function(resolve, reject)`.
#'
#' @return A promise object (see [then()]).
#'
#' @examples
#' p1 <- promise(function(resolve, reject) {
#'   later(function() resolve(runif(1)), delay = 2)
#' })
#' then(p1, print)
#' run_now(3)
#'
#' @export
promise <- function(action) {
  if (!is.function(action) || length(formals(action)) != 2) {
    stop("'action' must be a function with two arguments: `resolve`, `reject`")
  }

  p <- Promise()

  tryCatch(
    action(p$resolve, p$reject),
    error = function(e) {
      if (p$status() == "pending") {
        p$reject(e)
      } else {
        # Too late to do anything useful. Just notify.
        warning(e)
      }
    }
  )

  structure(
    list(
      then = p$then,
      catch = p$catch,
      finally = p$finally
    ),
    class = "promise",
    promise_impl = p
  )
}

#' @title Determine whether an object is a promise
#' @param x An R object to test.
#' @return `TRUE` if `x` is a promise object created by [promise()], `FALSE`
#'   otherwise.
#' @export
is.promise <- function(x) {
  inherits(x, "promise")
}

#' @title Format a promise
#' @param x A promise object.
#' @param ... Further arguments passed to or from other methods (currently
#'   unused).
#' @return A character string.
#' @keywords internal
#' @exportS3Method
format.promise <- function(x, ...) {
  attr(x, "promise_impl", exact = TRUE)$format()
}

#' @title Print a promise
#' @param x A promise object.
#' @param ... Further arguments passed to or from other methods (just for S3
#' complicance, unused)
#' @return `NULL`, invisibly. Called for its side effect of printing.
#' @keywords internal
#' @exportS3Method
print.promise <- function(x, ...) {
  cat(paste(format(x), collapse = "\n"), "\n", sep = "")
}

#' @title Access the results of a promise
#' 
#' @description Use `then()` to access the eventual result of a promise (or, if
#' the operation fails, the reason for that failure). The call to `then()` is
#' non-blocking: it returns immediately, and the return value is itself a
#' new promise.
#'
#' `catch()` is equivalent to `then()`, but without the `onFulfilled`
#' argument; it is typically used at the end of a promise chain to perform
#' error handling.
#'
#' `finally()` is similar to `then()`, but takes a single no-argument
#' function that is executed upon completion of the promise, regardless of
#' whether the result is success or failure. The return value of the
#' `onFinally` callback is ignored; an error thrown from it is propagated
#' forward into the returned promise.
#'
#' @param promise A promise object.
#' @param onFulfilled A function to be invoked if `promise` resolves
#'   successfully. Called with the resolved value, and optionally a second
#'   `.visible` argument indicating whether the value is
#'   [visible][base::invisible()].
#' @param onRejected A function taking a single `reason` argument, to be
#'   invoked if `promise` fails.
#'
#' @return A new promise.
#'
#' @export
then <- function(promise, onFulfilled = NULL, onRejected = NULL) {
  if (!is.promise(promise)) {
    stop("`promise` must be a promise object")
  }
  if (!is.null(onFulfilled) && !is.function(onFulfilled)) {
    stop("`onFulfilled` must be a function or `NULL`")
  }
  if (!is.null(onRejected) && !is.function(onRejected)) {
    stop("`onRejected` must be a function or `NULL`")
  }

  invisible(promise$then(onFulfilled = onFulfilled, onRejected = onRejected))
}

#' @rdname then
#' @export
catch <- function(promise, onRejected) {
  if (!is.promise(promise)) {
    stop("`promise` must be a promise object")
  }
  if (!is.function(onRejected)) {
    stop("`onRejected` must be a function")
  }

  invisible(promise$catch(onRejected))
}

#' @param onFinally A function with no arguments, called when `promise`
#'   either succeeds or fails.
#' @rdname then
#' @export
finally <- function(promise, onFinally) {
  if (!is.promise(promise)) {
    stop("`promise` must be a promise object")
  }
  if (!is.function(onFinally)) {
    stop("`onFinally` must be a function")
  }

  invisible(promise$finally(onFinally))
}

# nocov start
# tested implicitly

.onLoad <- function(libname, pkgname) {
  ensureInitialized()
  .globals$next_id <- 0L
  # Store a ref to the global loop so it doesn't get GC'd.
  .globals$global_loop <- create_loop(parent = NULL)
}

# nocov end

.globals <- new.env(parent = emptyenv())
# A registry of weak refs to loop handle objects. Given an ID number, we can
# get the corresponding loop handle. We use weak refs because we don't want
# this registry to keep the loop objects alive.
.loops <- new.env(parent = emptyenv())

#' @title Private event loops
#'
#' @description Normally, later uses a global event loop for scheduling and
#' running functions. However, in some cases, it is useful to create a
#' \emph{private} event loop to schedule and execute tasks without disturbing
#' the global event loop. For example, you might have asynchronous code that
#' queries a remote data source, but want to wait for a full back-and-forth
#' communication to complete before continuing in your code -- from the caller's
#' perspective, it should behave like synchronous code, and not do anything with
#' the global event loop (which could run code unrelated to your operation). To
#' do this, you would run your asynchronous code using a private event loop.
#'
#' \code{create_loop} creates and returns a handle to a private event loop,
#' which is useful when for scheduling tasks when you do not want to interfere
#' with the global event loop.
#'
#' \code{destroy_loop} destroys a private event loop.
#'
#' \code{exists_loop} reports whether an event loop exists -- that is, that it
#' has not been destroyed.
#'
#' \code{current_loop} returns the currently-active event loop. Any calls to
#' \code{\link{later}()} or \code{\link{run_now}()} will use the current loop by
#' default.
#'
#' \code{with_loop} evaluates an expression with a given event loop as the
#' currently-active loop.
#'
#' \code{with_temp_loop} creates an event loop, makes it the current loop, then
#' evaluates the given expression. Afterwards, the new event loop is destroyed.
#'
#' \code{global_loop} returns a handle to the global event loop.
#'
#' @param loop A handle to an event loop.
#' @param expr An expression to evaluate.
#' @param parent The parent event loop for the one being created. Whenever the
#'   parent loop runs, this loop will also automatically run, without having to
#'   manually call \code{\link{run_now}()} on this loop. If \code{NULL}, then
#'   this loop will not have a parent event loop that automatically runs it; the
#'   only way to run this loop will be by calling \code{\link{run_now}()} on this
#'   loop.
#' 
#' @rdname create_loop
#' 
#' @return An object of class `event_loop`.
#'
#' @export
create_loop <- function(parent = current_loop()) {
  id <- .globals$next_id
  .globals$next_id <- id + 1L

  if (is.null(parent)) {
    parent_id <- -1L
  } else if (inherits(parent, "event_loop")) {
    parent_id <- parent$id
  } else {
    stop("`parent` must be NULL or an event_loop object.")
  }
  createCallbackRegistry(id, parent_id)

  # Create the handle for the loop
  loop <- new.env(parent = emptyenv())
  class(loop) <- "event_loop"
  loop$id <- id
  lockBinding("id", loop)

  # Add a weak reference to the loop object in our registry.
  .loops[[as.character(id)]] <- new_weakref(loop)

  if (id != 0L) {
    # Inform the C++ layer that there are no more R references when the handle
    # is GC'd (unless it's the global loop.) The global loop handle never gets
    # GC'd under normal circumstances because .globals$global_loop refers to it.
    # However, if the package is unloaded it can get GC'd, and we don't want the
    # destroy_loop() finalizer to give an error message about not being able to
    # destroy the global loop.
    reg.finalizer(loop, notify_r_ref_deleted, onexit = TRUE)
  }

  loop
}

notify_r_ref_deleted <- function(loop) {
  res <- notifyRRefDeleted(loop$id)
  if (res) {
    rm(list = as.character(loop$id), envir = .loops)
  }
  invisible(res)
}

#' @rdname create_loop
#' @export
destroy_loop <- function(loop) {
  res <- deleteCallbackRegistry(loop$id)
  if (res) {
    rm(list = as.character(loop$id), envir = .loops)
  }
  invisible(res)
}

#' @rdname create_loop
#' @export
exists_loop <- function(loop) {
  existsCallbackRegistry(loop$id)
}

#' @rdname create_loop
#' @export
current_loop <- function() {
  id <- getCurrentRegistryId()
  loop_weakref <- .loops[[as.character(id)]]

  loop <- wref_key(loop_weakref)
  if (is.null(loop)) {
    stop("Current loop with id ", id, " not found.")
  }

  loop
}

#' @rdname create_loop
#' @export
with_temp_loop <- function(expr) {
  loop <- create_loop(parent = NULL)
  on.exit(destroy_loop(loop))

  with_loop(loop, expr)
}

#' @rdname create_loop
#' @export
with_loop <- function(loop, expr) {
  if (!exists_loop(loop)) {
    stop("loop has been destroyed!")
  }
  old_loop <- current_loop()
  if (!identical(loop, old_loop)) {
    on.exit(setCurrentRegistryId(old_loop$id))
    setCurrentRegistryId(loop$id)
  }

  force(expr)
}

#' @rdname create_loop
#' @export
global_loop <- function() {
  .globals$global_loop
}

#' @title Format an event loop
#' @param x An event loop object.
#' @param ... Further arguments passed to or from other methods (currently
#'   unused).
#' @return A character string.
#' @keywords internal
#' @exportS3Method
format.event_loop <- function(x, ...) {
  str <- paste0("<event loop> ID: ", x$id)
  if (!exists_loop(x)) {
    str <- paste(str, "(destroyed)")
  }
  str
}

#' @title Print an event loop
#' @param x An event loop object.
#' @param ... Further arguments passed to or from other methods (currently
#'   unused).
#' @return `NULL`, invisibly. Called for its side effect of printing.
#' @keywords internal
#' @exportS3Method
print.event_loop <- function(x, ...) {
  cat(format(x, ...))
}


#' @title Executes a function later
#'
#' @description Schedule an R function or formula to run after a specified
#' period of time. Similar to JavaScript's `setTimeout` function. Like
#' JavaScript, R is single-threaded so there's no guarantee that the operation
#' will run exactly at the requested time, only that at least that much time
#' will elapse.
#'
#' The mechanism used by this package is inspired by Simon Urbanek's
#' [background](https://github.com/s-u/background) package and similar code in
#' Rhttpd.
#'
#' To avoid bugs due to reentrancy, by default, scheduled operations only run
#' when there is no other R code present on the execution stack; i.e., when R is
#' sitting at the top-level prompt. You can force past-due operations to run at
#' a time of your choosing by calling [run_now()].
#'
#' Error handling is not particularly well-defined and may change in the future.
#' options(error=browser) should work and errors in `func` should generally not
#' crash the R process, but not much else can be said about it at this point.
#' If you must have specific behavior occur in the face of errors, put error
#' handling logic inside of `func`.
#'
#' @param func A function or one-sided formula (e.g. `~ .x + 1`, using `.` or
#'   `.x` as the argument pronoun).
#' @param delay Number of seconds in the future to delay execution. There is no
#'   guarantee that the function will be executed at the desired time, but it
#'   should not execute earlier.
#' @param loop A handle to an event loop. Defaults to the currently-active loop.
#'
#' @return A function, which, if invoked, will cancel the callback. The
#'   function will return \code{TRUE} if the callback was successfully
#'   cancelled and \code{FALSE} if not (this occurs if the callback has
#'   executed or has been cancelled already).
#'
#' @examples
#' # Example of formula style
#' later(~ cat("Hello from the past\n"), 3)
#'
#' # Example of function style
#' later(function() {
#'   print(summary(cars))
#' }, 2)
#'
#' @export
later <- function(func, delay = 0, loop = current_loop()) {
  if (!is.function(func)) {
    func <- as_function(func)
  }
  id <- execLater(func, delay, loop$id)

  invisible(create_canceller(id, loop$id))
}

# Converts a one-sided formula (e.g. `~ .x + 1`) into a function, using `.`
# and `.x` as the pronoun for the first argument. Plain functions are
# returned unchanged. This is a small base-R stand-in for
# `rlang::as_function()`, used so that `later`/`later_fd` don't need to
# depend on `rlang`.
as_function <- function(x) {
  if (is.function(x)) {
    return(x)
  }
  if (inherits(x, "formula")) {
    if (length(x) != 2L) {
      stop("`func` must be a function or a one-sided formula, e.g. `~ .x + 1`.")
    }
    rhs <- x[[2L]]
    env <- attr(x, ".Environment")
    f <- function(...) {
      dots <- list(...)
      . <- if (length(dots) >= 1L) dots[[1L]] else NULL
      .x <- .
      eval(rhs, envir = list(. = ., .x = .x), enclos = env)
    }
    return(f)
  }
  stop("`func` must be a function or a formula.")
}

# Returns a function that will cancel a callback with the given ID. If the
# callback has already been executed or canceled, then the function has no
# effect.
create_canceller <- function(id, loop_id) {
  force(id)
  force(loop_id)
  function() {
    invisible(cancel(id, loop_id))
  }
}

#' @title Executes a function when a file descriptor is ready
#'
#' @description Schedule an R function or formula to run after an indeterminate
#' amount of time when file descriptors are ready for reading or writing,
#' subject to an optional timeout.
#'
#' On the occasion the system-level `poll` (on Windows `WSAPoll`) returns an
#' error, the callback will be made on a vector of all `NA`s. This is
#' indistinguishable from a case where the `poll` succeeds but there are error
#' conditions pending against each file descriptor. If no file descriptors are
#' supplied, the callback is scheduled for immediate execution and made on the
#' empty logical vector `logical(0)`.
#'
#' @param func A function that takes a single argument, a logical vector that
#'   indicates which file descriptors are ready (a concatenation of `readfds`,
#'   `writefds` and `exceptfds`). This may be all `FALSE` if the
#'   `timeout` argument is non-`Inf`. File descriptors with error conditions
#'   pending are represented as `NA`, as are invalid file descriptors such as
#'   those already closed.
#' @param readfds Integer vector of file descriptors, or Windows SOCKETs, to
#'   monitor for being ready to read.
#' @param writefds Integer vector of file descriptors, or Windows SOCKETs, to
#'   monitor being ready to write.
#' @param exceptfds Integer vector of file descriptors, or Windows SOCKETs, to
#'   monitor for error conditions pending.
#' @param timeout Number of seconds to wait before giving up, and calling `func`
#'   with all `FALSE`. The default `Inf` implies waiting indefinitely.
#'   Specifying `0` will check once without blocking, and supplying a negative
#'   value defaults to a timeout of 1s.
#' @param loop A handle to an event loop. Defaults to the currently-active loop.
#'
#' @inherit later return note
#'
#' @examples
#' \dontrun{
#' # Unix-only example
#' 
#' # Use the base R 'parallel' package to fork child processes and obtain
#' # real, pollable file descriptors. A child's `fd` becomes ready for
#' # reading once the child finishes running.
#'
#' # 1. timeout: prints FALSE, FALSE
#' job1 <- parallel::mcparallel({
#'   Sys.sleep(1)
#'   TRUE
#' })
#' job2 <- parallel::mcparallel({
#'   Sys.sleep(1)
#'   TRUE
#' })
#' fd1 <- job1$fd[1]
#' fd2 <- job2$fd[1]
#' later_fd(print, c(fd1, fd2), timeout = 0.1)
#' Sys.sleep(0.2)
#' run_now()
#'
#' # 2. fd1 ready: prints TRUE, FALSE
#' job1 <- parallel::mcparallel(TRUE)
#' fd1 <- job1$fd[1]
#' Sys.sleep(0.1)
#' later_fd(print, c(fd1, fd2), timeout = 1)
#' Sys.sleep(0.1)
#' run_now()
#'
#' # 3. both ready: prints TRUE, TRUE
#' job2 <- parallel::mcparallel(TRUE)
#' fd2 <- job2$fd[1]
#' Sys.sleep(0.1)
#' later_fd(print, c(fd1, fd2), timeout = 1)
#' Sys.sleep(0.1)
#' run_now()
#'
#' # 4. fd2 ready: prints FALSE, TRUE
#' parallel::mccollect(job1)
#' job1 <- parallel::mcparallel({
#'   Sys.sleep(1)
#'   TRUE
#' })
#' fd1 <- job1$fd[1]
#' later_fd(print, c(fd1, fd2), timeout = 1)
#' Sys.sleep(0.1)
#' run_now()
#'
#' # 5. fds invalid: prints NA, NA
#' parallel::mccollect(job1)
#' parallel::mccollect(job2)
#' later_fd(print, c(fd1, fd2), timeout = 0)
#' Sys.sleep(0.1)
#' run_now()
#' }
#'
#' @return A delayed function evaluation.
#'
#' @export
later_fd <- function(
  func,
  readfds = integer(),
  writefds = integer(),
  exceptfds = integer(),
  timeout = Inf,
  loop = current_loop()
) {
  if (!is.function(func)) {
    func <- as_function(func)
  }
  xptr <- execLater_fd(func, readfds, writefds, exceptfds, timeout, loop$id)

  invisible(create_fd_canceller(xptr))
}

# Returns a function that will cancel a callback with the given external
# pointer. If the callback has already been executed or canceled, then the
# function has no effect.
create_fd_canceller <- function(xptr) {
  force(xptr)
  function() {
    invisible(fd_cancel(xptr))
  }
}

#' @title Execute scheduled operations
#'
#' @description Normally, operations scheduled with [later()] will not execute
#' unless/until no other R code is on the stack (i.e. at the top-level). If you
#' need to run blocking R code for a long time and want to allow scheduled
#' operations to run at well-defined points of your own operation, you can call
#' `run_now()` at those points and any operations that are due to run will do
#' so.
#'
#' If one of the callbacks throws an error, the error will _not_ be caught, and
#' subsequent callbacks will not be executed (until `run_now()` is called again,
#' or control returns to the R prompt). You must use your own
#' [tryCatch][base::conditions] if you want to handle errors.
#'
#' @param timeoutSecs Wait (block) for up to this number of seconds waiting for
#'   an operation to be ready to run. If `0`, then return immediately if there
#'   are no operations that are ready to run. If `Inf` or negative, then wait as
#'   long as it takes (if none are scheduled, then this will block forever).
#' @param all If `FALSE`, `run_now()` will execute at most one scheduled
#'   operation (instead of all eligible operations). This can be useful in cases
#'   where you want to interleave scheduled operations with your own logic.
#' @param loop A handle to an event loop. Defaults to the currently-active loop.
#'
#' @return A logical indicating whether any callbacks were actually run.
#'
#' @export
run_now <- function(timeoutSecs = 0L, all = TRUE, loop = current_loop()) {
  invisible(execCallbacks(timeoutSecs, all, loop$id))
}

#' @title Check if later loop is empty
#' @description Returns true if there are currently no callbacks that are scheduled to
#' execute in the present or future.
#' @inheritParams create_loop
#' @keywords internal
#' @return A logical value.
#' @export
loop_empty <- function(loop = current_loop()) {
  idle(loop$id)
}

#' @title Relative time to next scheduled operation
#' @description Returns the duration between now and the earliest operation that
#' is currently scheduled, in seconds. If the operation is in the past, the
#' value will be negative. If no operation is currently scheduled, the value
#' will be `Inf`.
#' @inheritParams create_loop
#' @return A numeric value.
#' @export
next_op_secs <- function(loop = current_loop()) {
  nextOpSecs(loop$id)
}

#' @title Get the contents of an event loop, as a list
#' @description This function is for debugging only.
#' @keywords internal
list_queue <- function(loop = current_loop()) {
  list_queue_(loop$id)
}

# tests that compile snippets were moved to side pkg latertest

jitter <- 0.017 * 2 # Compensate for imprecision in system timer

source(system.file("tinytest", "helper.R", package = "later2"))

local({
  # run_now waits and returns FALSE if no tasks ----

  x <- system.time({
    result <- run_now(0.5)
  })
  expect_true(as.numeric(x[["elapsed"]]) >= 0.5 - jitter)
  expect_identical(result, FALSE)

  x <- system.time({
    result <- run_now(3)
  })
  expect_true(as.numeric(x[["elapsed"]]) >= 3 - jitter)
  expect_identical(result, FALSE)
})

local({
  # run_now returns immediately after executing a task ----

  x <- system.time({
    later(~ {}, 0)
    result <- run_now(2)
  })
  expect_true(as.numeric(x[["elapsed"]]) < 0.25)
  expect_identical(result, TRUE)
})

local({
  # run_now executes all scheduled tasks, not just one ----

  later(~ {}, 0)
  later(~ {}, 0)
  result1 <- run_now()
  result2 <- run_now()
  expect_identical(result1, TRUE)
  expect_identical(result2, FALSE)
})

local({
  # run_now executes just one scheduled task, if requested ----

  result1 <- run_now()
  expect_identical(result1, FALSE)

  later(~ {}, 0)
  later(~ {}, 0)

  result2 <- run_now(all = FALSE)
  expect_identical(result2, TRUE)

  result3 <- run_now(all = FALSE)
  expect_identical(result3, TRUE)

  result4 <- run_now()
  expect_identical(result4, FALSE)
})

local({
  # run_now doesn't go past a failed task ----

  later(~ stop("boom"), 0)
  later(~ {}, 0)
  expect_error(run_now())
  expect_true(run_now())
})


local({
  # Callbacks cannot affect the caller ----

  # This is based on a pattern used in the callCC function. Normally, simply
  # touching `throw` will cause the expression to be evaluated and f() to return
  # early. (This test does not involve later.)
  f <- function() {
    delayedAssign("throw", return(100))
    g <- function() {
      throw
    }
    g()
    return(200)
  }
  expect_equal(f(), 100)

  # When later runs callbacks, it wraps the call in R_ToplevelExec(), which
  # creates a boundary on the call stack that the early return can't cross.
  f <- function() {
    delayedAssign("throw", return(100))
    later(function() {
      throw
    })

    run_now(1)
    return(200)
  }
  # jcheng 2024-10-24: Apparently this works now, maybe because having
  # RCPP_USING_UNWIND_PROTECT means we don't need to use R_ToplevelExec to call
  # callbacks?
  # expect_error(f())
  expect_identical(f(), 100)

  # In this case, f() should return normally, and then when g() causes later to
  # run the callback with `throw`, it should be an error -- there's no function
  # to return from because it (f()) already returned.
  f <- function() {
    delayedAssign("throw", return(100))
    later(function() {
      throw
    })
    return(200)
  }
  g <- function() {
    run_now(2)
  }
  expect_equal(f(), 200)
  expect_error(g())
})


local({
  # interrupt and exception handling, R ----

  # =======================================================
  # Errors and interrupts in R callbacks
  # =======================================================

  # R error
  error_obj <- FALSE
  tryCatch(
    {
      later(function() {
        stop("oopsie")
      })
      run_now()
    },
    error = function(e) {
      error_obj <<- e
    }
  )
  expect_true(grepl("oopsie", error_obj$message))

  # interrupt
  interrupted <- FALSE
  tryCatch(
    {
      later(function() {
        simulate_interrupt()
        Sys.sleep(100)
      })
      run_now()
    },
    interrupt = function(e) {
      interrupted <<- TRUE
    }
  )
  expect_true(interrupted)
})

local({
  # Cancelling callbacks ----

  # Cancel with zero delay
  x <- 0
  cancel <- later2::later(function() {
    x <<- x + 1
  })
  later2::later(function() {
    x <<- x + 2
  })
  expect_true(cancel())
  later2::run_now()
  expect_identical(x, 2)

  # Cancel with zero delay
  x <- 0
  cancel <- later2::later(
    function() {
      x <<- x + 1
    },
    1
  )
  run_now(0.25)
  expect_true(cancel())
  run_now(1)
  expect_identical(x, 0)

  # Make sure a cancelled callback doesn't interfere with others
  x <- 0
  later2::later(
    function() {
      x <<- x + 1
    },
    1
  )
  cancel <- later2::later(
    function() {
      x <<- x + 2
    },
    0.5
  )
  later2::run_now()
  expect_true(cancel())
  run_now(2)
  expect_identical(x, 1)
})


local({
  # Cancelled functions will be GC'd ----

  x <- 0
  cancel <- later2::later(
    local({
      reg.finalizer(environment(), function(e) x <<- x + 1)
      function() message("foo")
    })
  )
  expect_true(cancel())
  gc()
  expect_identical(x, 1)
})


local({
  # Cancelling executed or cancelled callbacks has no effect ----

  # Cancelling an executed callback
  x <- 0
  cancel <- later2::later(function() {
    x <<- x + 1
  })
  later2::run_now()
  expect_false(cancel())
  later2::run_now()
  expect_identical(x, 1)

  # Cancelling twice
  x <- 0
  cancel <- later2::later(function() {
    x <<- x + 1
  })
  expect_true(cancel())
  expect_false(cancel())
  later2::run_now()
  expect_identical(x, 0)
})


local({
  # Cancelling callbacks on temporary event loops ----

  with_temp_loop({
    # Cancelling an executed callback
    x <- 0
    cancel <- later2::later(function() {
      x <<- x + 1
    })
    later2::run_now()
    expect_false(cancel())
    later2::run_now()
    expect_identical(x, 1)
  })

  with_temp_loop({
    # Cancelling twice
    x <- 0
    cancel <- later2::later(function() {
      x <<- x + 1
    })
    expect_true(cancel())
    expect_false(cancel())
    later2::run_now()
    expect_identical(x, 0)
  })

  with_temp_loop({
    # Make sure a cancelled callback doesn't interfere with others
    x <- 0
    later2::later(
      function() {
        x <<- x + 1
      },
      1
    )
    cancel <- later2::later(
      function() {
        x <<- x + 2
      },
      0.5
    )
    later2::run_now()
    expect_true(cancel())
    run_now(2)
    expect_identical(x, 1)
  })

  # Canceling after an event loop handle has been destroyed: the underlying
  # data structure (in C++) will be deleted, along with the callbacks. This is
  # true because the loop does not have a parent.
  cancel <- NULL
  x <- 0
  with_temp_loop({
    cancel <- later2::later(function() {
      x <<- x + 1
    })
  })
  expect_false(cancel())
  expect_identical(x, 0)
})


local({
  # Cancelling callbacks on persistent private loops without parent ----

  l1 <- create_loop(parent = NULL)
  l2 <- create_loop(parent = NULL)

  # Cancel from outside with_loop
  cancel <- NULL
  x <- 0
  with_loop(l1, {
    cancel <- later2::later(function() {
      x <<- x + 1
    })
  })
  expect_true(cancel())
  expect_false(cancel())
  with_loop(l1, later2::run_now())
  expect_false(cancel())
  expect_identical(x, 0)

  # Make sure it doesn't interfere with other event loops
  with_loop(l1, {
    cancel <- later2::later(function() {
      x <<- x + 1
    })
  })
  with_loop(l2, {
    later2::later(function() {
      x <<- x + 2
    })
  })
  later2::later(function() {
    x <<- x + 4
  })
  expect_true(cancel())
  with_loop(l1, later2::run_now())
  with_loop(l2, later2::run_now())
  later2::run_now()
  expect_identical(x, 6)

  # Cancelling on an explicitly destroyed loop returns FALSE
  l3 <- create_loop(parent = NULL)
  cancel <- NULL
  x <- 0
  with_loop(l3, {
    cancel <- later2::later(function() {
      x <<- x + 1
    })
  })
  destroy_loop(l3)
  expect_false(cancel())
  expect_identical(x, 0)
})

local({
  # Cancelling callbacks on persistent private loops with parent ----

  # If the loop handle is GC'd but the loop _does have_ a parent, then the
  # underlying objects will not be destroyed right away, so the cancel() will
  # work.
  cancel <- NULL
  x <- 0
  local({
    l1 <- create_loop(parent = current_loop())
    cancel <<- later2::later(
      function() {
        x <<- x + 1
      },
      loop = l1
    )
  })
  expect_true(cancel())
  expect_false(cancel())
  expect_identical(x, 0)
})

local({
  # A canceler will not keep loop alive ----

  l <- create_loop(parent = NULL)
  finalized <- FALSE

  reg.finalizer(l, function(x) finalized <<- TRUE)
  cancel <- later2::later(function() 1, loop = l)
  rm(l)
  gc()
  expect_true(finalized)
})

local({
  # Canceling a callback from another a callback ----

  # Canceling a callback from another callback should work. Additionally, the
  # altered ordering of callbacks shouldn't prevent other callbacks from
  # running. In this test, #1 cancels 2 and 3, but we still expect 4 to run. If
  # we used the wrong algorithm for traversing the queue and canceling
  # callbacks, it would be possible for the cancellation of 2 and 3 to cause 4
  # to not run. This test ensures that we do it the right way.
  ran_2 <- FALSE
  ran_3 <- FALSE
  ran_4 <- FALSE
  with_temp_loop({
    cancel_1 <- later2::later(function() {
      cancel_2()
      cancel_3()
    })
    cancel_2 <- later2::later(function() {
      ran_2 <<- TRUE
    })
    cancel_3 <- later2::later(function() {
      ran_3 <<- TRUE
    })
    later2::later(function() {
      ran_4 <<- TRUE
    })
    later2::run_now()
  })

  expect_false(ran_2)
  expect_false(ran_3)
  expect_true(ran_4)
})

source(system.file("tinytest", "helper.R", package = "later2"))

local({
  # Private event loop changes current_loop() ----

  expect_identical(current_loop(), global_loop())

  with_temp_loop({
    expect_false(identical(current_loop(), global_loop()))
  })
})

local({
  # Private event loop runs only its own tasks ----

  x <- 0
  later(
    ~ {
      x <<- 1
    },
    0
  )
  with_temp_loop({
    expect_true(loop_empty())

    later(
      ~ {
        x <<- 2
      }
    )
    later2::run_now()

    expect_identical(x, 2)

    run_now(loop = global_loop())
    expect_identical(x, 1)
  })
})

local({
  # Private event loops ----

  l <- create_loop(parent = NULL)
  x <- 0

  expect_true(exists_loop(l))

  with_loop(l, {
    later(function() x <<- x + 1)
    later2::run_now()
  })
  expect_equal(x, 1)

  with_loop(l, {
    later(function() x <<- x + 1)
    later2::run_now()

    later(function() x <<- x + 1)
    later(function() x <<- x + 1)
  })

  expect_equal(x, 2)

  with_loop(l, later2::run_now())
  expect_equal(x, 4)

  destroy_loop(l)
  expect_false(exists_loop(l))

  # Can't run later-y things with destroyed loop
  expect_error(with_loop(l, later(function() message("foo"))))
  expect_error(run_now(loop = l))

  # GC with functions in destroyed loops, even if callback isn't executed.
  l <- create_loop(parent = NULL)
  x <- 0
  gc()
  with_loop(l, {
    later(
      local({
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() message("foo")
      })
    )
  })
  gc()
  expect_identical(x, 0)

  destroy_loop(l)
  gc()
  expect_identical(x, 1)

  # A GC'd loop object will cause its queue to be deleted, which will allow GC
  # of any resources
  l <- create_loop(parent = NULL)
  x <- 0
  gc()
  with_loop(l, {
    later(
      local({
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() message("foo")
      })
    )
  })
  gc()
  expect_identical(x, 0)

  # Delete the reference to the loop, and GC. This causes the queue to be
  # deleted, which removes references to items in the queue. However, the items
  # in the queue won't be GC'd yet. (At least not as of R 3.5.2.)
  rm(l)
  gc()
  expect_identical(x, 0)

  # A second GC triggers the finalizer for an item that was in the queue.
  gc()
  expect_identical(x, 1)

  # Can't destroy global loop
  expect_error(destroy_loop(global_loop()))
})

local({
  # Temporary event loops ----

  l <- NULL
  x <- 0
  with_temp_loop({
    l <- current_loop()
    later(function() x <<- x + 1)
    later2::run_now()
  })

  expect_false(exists_loop(l))
  expect_error(
    with_loop(l, {
      later(function() x <<- x + 1)
      later2::run_now()
    })
  )

  # Test GC
  # Make sure that items captured in later callbacks are GC'd after the callback
  # is executed.
  x <- 0
  with_temp_loop({
    later(
      local({
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() 1
      })
    )
    gc()
    later2::run_now()
  })
  expect_identical(x, 0)
  gc()
  expect_identical(x, 1)

  # Test that objects are GC'd after loop is destroyed, even if callback hasn't
  # been executed.
  x <- 0
  with_temp_loop({
    later(
      local({
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() 1
      })
    )
    later2::run_now()

    later(
      local({
        e <- environment()
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() 1
      })
    )
    gc()
  })
  expect_identical(x, 1)
  gc()
  expect_identical(x, 2)
})

local({
  # Destroying loop and loop ID ----

  l <- create_loop()
  expect_true(is.integer(l$id))
  expect_true(destroy_loop(l))
  expect_false(exists_loop(l))

  # Should return false on subsequent calls to destroy_loop()
  expect_false(destroy_loop(l))
  # Destroying a second time shouldn't cause warnings.
  expect_silent(destroy_loop(l))
})

local({
  # Can't destroy current loop ----
  errored <- FALSE
  with_temp_loop({
    later(function() {
      # We can't do expect_error in a later() callback, so use a tryCatch
      # instead to check that an error occurs.
      tryCatch(
        destroy_loop(current_loop()),
        error = function(e) {
          errored <<- TRUE
        }
      )
    })
    later2::run_now()
  })

  expect_true(errored)
})

local({
  # Can't GC current loop ----

  collected <- FALSE
  l <- create_loop()
  reg.finalizer(l, function(x) {
    collected <<- TRUE
  })
  with_loop(l, {
    rm(l, inherits = TRUE)
    gc()
    gc()
  })
  expect_false(collected)
  gc()
  expect_true(collected)
})


local({
  # When auto-running a child loop, it will be reported as current_loop() ----

  l <- create_loop(parent = global_loop())
  x <- NULL
  later(
    function() {
      x <<- current_loop()
    },
    loop = l
  )
  run_now(loop = global_loop())
  expect_identical(x, l)
})

local({
  # CallbackRegistry exists until its callbacks are run ----

  # If the R loop handle object is GC'd, it doesn't necessarily destroy the
  # underlying CallbackRegistry (in C++). The CallbackRegistry is only destroyed
  # when the R loop handle is GC'd AND the CallbackRegistry contains no more
  # callbacks.
  x <- 0
  callback <- function() {
    x <<- x + 1
  }
  local({
    l <- create_loop()
    later(callback, loop = l)
  })
  gc()
  later2::run_now()
  expect_identical(x, 1)
})

local({
  # Auto-running grandchildren loops
  l1_ran <- FALSE
  l11_ran <- FALSE
  l12_ran <- FALSE
  l13_ran <- FALSE
  l2_ran <- FALSE
  l21_ran <- FALSE
  l22_ran <- FALSE
  l23_ran <- FALSE

  l1 <- create_loop()
  l2 <- create_loop(parent = NULL)

  # l1 should auto-run, along with l11 and l12. l13 should not, because it has
  # no parent.
  with_loop(l1, {
    later(function() l1_ran <<- TRUE)
    l11 <- create_loop()
    l12 <- create_loop()
    l13 <- create_loop(parent = NULL)
    later(function() l11_ran <<- TRUE, loop = l11)
    later(function() l12_ran <<- TRUE, loop = l12)
    later(function() l13_ran <<- TRUE, loop = l13)
  })

  # None of these should auto-run, because l2 has no parent.
  with_loop(l2, {
    later(function() l2_ran <<- TRUE)
    l21 <- create_loop()
    l22 <- create_loop()
    l23 <- create_loop(parent = NULL)
    later(function() l21_ran <<- TRUE, loop = l21)
    later(function() l22_ran <<- TRUE, loop = l22)
    later(function() l23_ran <<- TRUE, loop = l23)
  })

  later2::run_now()
  expect_true(l1_ran)
  expect_true(l11_ran)
  expect_true(l12_ran)
  expect_false(l13_ran)
  expect_false(l2_ran)
  expect_false(l21_ran)
  expect_false(l22_ran)
  expect_false(l23_ran)
})

local({
  # Grandchildren loops whose parent is destroyed should not autorun ----

  l_ran <- 0
  l1_ran <- 0
  l <- create_loop()

  with_loop(l, {
    later(function() l_ran <<- l_ran + 1)
    l1 <- create_loop()
    later(function() l1_ran <<- l1_ran + 1, loop = l1)
  })

  later2:::notify_r_ref_deleted(l)
  later2::run_now()
  # l will run, because the underlying registry exists until empty. It also
  # causes l1 to run.
  expect_identical(l_ran, 1)
  expect_identical(l1_ran, 1)
  expect_false(exists_loop(l))
  # l1 should still exist because we still have a reference to it.
  expect_true(exists_loop(l1))

  # Schedule another function that we don't expect to actually run.
  # Use finalizer to keep
  l1_finalized <- FALSE
  later(
    local({
      reg.finalizer(environment(), function(e) l1_finalized <<- TRUE)
      function() l1_ran <<- l1_ran + 1
    }),
    loop = l1
  )
  later2::run_now()
  # l1 won't run again
  expect_identical(l1_ran, 1)
  expect_true(exists_loop(l1))
  expect_false(l1_finalized)
  # If the reference is lost (like when the loop handle is GC'd) l1 will take
  # effect immediately.
  expect_true(later2:::notify_r_ref_deleted(l1))
  expect_false(exists_loop(l1))
  gc() # Make the finalizer run
  expect_true(l1_finalized)
})


local({
  # Removing parent loop allows loop to be deleted ----

  # Create parent loop, then create a child loop, then add a finalizer to a
  # callback (actually, the env for the callback) in the child loop.
  l <- create_loop()
  l1 <- create_loop(parent = l)

  x <- 0
  with_loop(l1, {
    later(
      local({
        reg.finalizer(environment(), function(e) x <<- x + 1)
        function() NULL
      })
    )
  })

  # Removing the ref to the child should NOT cause the finalizer to run -- the
  # loop won't actually be destroyed because it (A) has a parent AND (B) has a
  # callback. notify_r_ref_deleted(l1)
  rm(l1)
  gc()
  gc()
  expect_identical(x, 0)

  # If we destroy the parent loop, then the finalizer will be called, because
  # even though the child loop has a callback, it no longer has a parent.
  # Because both the handle and its parent have been GC'd, there's no way to run
  # callbacks in the child, so the internal representation of the child loop can
  # be deleted, along with all the callbacks it contains.
  rm(l)
  # Use 2 GC's because the first causes the loops to be GC'd; the second causes
  # the function that was queued in a loop to be GC'd.
  gc()
  gc()
  expect_identical(x, 1)
})

local({
  # Interrupt while running in private loop won't result in stuck loop ----

  l <- create_loop()
  later(
    function() {
      simulate_interrupt()
    },
    loop = l
  )
  tryCatch(
    {
      run_now(loop = l)
    },
    interrupt = function(e) NULL
  )
  expect_identical(current_loop(), global_loop())

  tryCatch(
    {
      with_loop(l, {
        simulate_interrupt()
      })
    },
    interrupt = function(e) NULL
  )
  expect_identical(current_loop(), global_loop())
})

local({
  # later2:::list_queue ----

  l <- create_loop(parent = NULL)
  q <- NULL
  f <- function() 1 # A dummy function

  with_loop(l, {
    later(f)
    q <- later2:::list_queue()
  })
  expect_equal(length(q), 1)
  expect_identical(q[[1]]$callback, f)

  with_loop(l, {
    later2::run_now()
    q <- later2:::list_queue()
  })
  expect_equal(length(q), 0)

  with_loop(l, {
    later(f)
    later(f)
    later(sum)
    q <- later2:::list_queue()
  })
  expect_equal(length(q), 3)
  expect_identical(q[[1]]$callback, f)
  expect_identical(q[[2]]$callback, f)
  expect_identical(q[[3]]$callback, sum)

  # Empty the queue by calling run now. Also test calling later2:::list_queue by passing
  # in a loop handle.
  with_loop(l, later2::run_now())
  q <- later2:::list_queue(l)
  expect_equal(length(q), 0)

  destroy_loop(l)
  expect_error(later2:::list_queue(l))
})

local({
  # next_op_secs works ----

  loop <- with_temp_loop({
    expect_identical(next_op_secs(), Inf)

    later(function() {}, 0)
    expect_true(next_op_secs() <= 0)
    later2::run_now()

    later(function() {}, 0.1)
    expect_true(next_op_secs() <= 0.1)
    run_now(2)

    current_loop()
  })
  expect_error(next_op_secs(loop))
})

local({
  # parameter validation works ----

  loop <- create_loop(parent = NULL)
  expect_error(with_loop(loop, destroy_loop(loop)))
  expect_true(destroy_loop(loop))
  expect_false(destroy_loop(loop))
  expect_error(with_loop(loop, {}))
  expect_error(loop_empty(loop))
  expect_error(create_loop(parent = "invalid"))
})

local({
  # print.event_loop works correctly ----

  loop <- create_loop(parent = NULL)

  output <- capture.output(print(loop))
  expect_match(output, "<event loop> ID: [0-9]+")

  destroy_loop(loop)
  output_destroyed <- capture.output(print(loop))
  expect_match(output_destroyed, "\\(destroyed\\)")
})

local({
  # esoteric error handlers ----
  
  loop <- create_loop(parent = NULL)
  expect_error(later2:::notify_r_ref_deleted(global_loop()))
  expect_error(with_loop(loop, later2:::notify_r_ref_deleted(loop)))
  expect_error({
    with_loop(loop, {
      .loops[[as.character(loop$id)]] <- NULL
      current_loop()
    })
  })
})

local({
  # later_fd ----

  # Uses the 'parallel' package (part of every base R installation) to
  # obtain real, pollable file descriptors instead of depending on an
  # external socket library. `parallel::mcparallel()` forks a child process
  # and hands back a pipe file descriptor (`$fd[1]`) in the parent that
  # becomes ready for reading once the child produces its result.
  
  # This is unix-only (fork-based), so the test is skipped elsewhere. It's
  # also skipped on CRAN: it races the fd-polling background thread's
  # internal timing (capped at ~1s per poll(), see src/fd.cpp) against
  # run_now()'s wall-clock budget with little to no margin, which is prone
  # to spurious failures on CRAN's loaded/virtualized check machines.
  if (.Platform$OS.type != "unix") { return(NULL) }
  if (!identical(Sys.getenv("NOT_CRAN"), "true")) { return(NULL) }

  jobs <- list()

  # A job whose fd becomes ready almost immediately.
  make_ready_fd <- function() {
    job <- parallel::mcparallel(TRUE)
    jobs[[length(jobs) + 1]] <<- job
    Sys.sleep(0.3)
    job
  }

  # A job whose fd stays "not ready" for the lifetime of the test.
  make_pending_fd <- function() {
    job <- parallel::mcparallel({ Sys.sleep(5); TRUE })
    jobs[[length(jobs) + 1]] <<- job
    job
  }

  on.exit({
    for (job in jobs) {
      tryCatch(tools::pskill(job$pid), error = function(e) NULL)
      suppressWarnings(parallel::mccollect(job, wait = TRUE, timeout = 1))
    }
  })

  result <- NULL
  callback <- function(x) result <<- x

  # timeout (both fds pending)
  job1 <- make_pending_fd()
  job2 <- make_pending_fd()
  fd1 <- job1$fd[1]
  fd2 <- job2$fd[1]
  later_fd(callback, c(fd1, fd2), timeout = 0)
  run_now(1)
  expect_equal(result, c(FALSE, FALSE))
  later_fd(callback, c(fd1, fd2), exceptfds = c(fd1, fd2), timeout = 0)
  run_now(1)
  expect_equal(result, c(FALSE, FALSE, FALSE, FALSE))

  # cancellation
  result <- NULL
  cancel <- later_fd(callback, c(fd1, fd2), timeout = 0.2)
  expect_equal(typeof(cancel), "closure")
  expect_true(cancel())
  Sys.sleep(0.25)
  expect_false(cancel())
  vis <- withVisible(cancel())
  expect_false(vis$visible)
  later2::run_now()
  expect_null(result)

  # timeout (> 1 loop)
  later_fd(callback, c(fd1, fd2), timeout = 1.1)
  run_now(1.8)
  expect_equal(result, c(FALSE, FALSE))

  # fd1 ready, fd2 pending
  job1 <- make_ready_fd()
  fd1 <- job1$fd[1]
  later_fd(callback, c(fd1, fd2), timeout = 0.9)
  run_now(1.5)
  expect_equal(result, c(TRUE, FALSE))

  # both fd1, fd2 ready
  job2 <- make_ready_fd()
  fd2 <- job2$fd[1]
  later_fd(callback, c(fd1, fd2), timeout = 1)
  run_now(1.5)
  expect_equal(result, c(TRUE, TRUE))

  # no exceptions (fds still open and their processes still running, so
  # there's no hang-up condition; note we can't reuse fd1/fd2 here, since by
  # now their child processes have long since finished and exited, and the
  # closed write end would trigger a hang-up, which is correctly reported as
  # an exception/NA rather than FALSE)
  jobA <- make_pending_fd()
  jobB <- make_pending_fd()
  fdA <- jobA$fd[1]
  fdB <- jobB$fd[1]
  later_fd(callback, c(fdA, fdB), exceptfds = c(fdA, fdB), timeout = -0.1)
  run_now(1.5)
  expect_equal(result, c(FALSE, FALSE, FALSE, FALSE))

  # fd1 not ready, fd2 ready
  job1 <- make_pending_fd()
  fd1 <- job1$fd[1]
  later_fd(callback, c(fd1, fd2), timeout = 1L)
  run_now(1.5)
  expect_equal(result, c(FALSE, TRUE))

  # fd2 invalid (already collected/closed)
  suppressWarnings(parallel::mccollect(job2, wait = TRUE, timeout = 1))
  later_fd(callback, c(fd1, fd2), exceptfds = c(fd1, fd2), timeout = 0.1)
  run_now(1)
  expect_length(result, 4L)

  # both fd1, fd2 invalid
  suppressWarnings(parallel::mccollect(job1, wait = TRUE, timeout = 1))
  later_fd(callback, c(fd1, fd2), c(fd1, fd2), timeout = 0)
  run_now(1)
  expect_equal(result, c(NA, NA, NA, NA))

  # no fds supplied
  later_fd(callback, timeout = -1)
  run_now(1.5)
  expect_equal(result, logical())
})

local({
  # loop_empty() reflects later_fd callbacks ----

  # See the comment in the previous local() block: skipped on CRAN because
  # it's fork-based and races background-thread timing against wall clock.
  if (.Platform$OS.type != "unix") { return(NULL) }
  if (!identical(Sys.getenv("NOT_CRAN"), "true")) { return(NULL) }

  job <- parallel::mcparallel({ Sys.sleep(5); TRUE })
  on.exit({
    tryCatch(tools::pskill(job$pid), error = function(e) NULL)
    suppressWarnings(parallel::mccollect(job, wait = TRUE, timeout = 1))
  })

  fd1 <- job$fd[1]

  expect_true(loop_empty())

  cancel <- later_fd(~ {}, fd1)
  expect_false(loop_empty())
  cancel()
  Sys.sleep(1.25) # check for cancellation happens every ~1 sec
  expect_true(loop_empty())

  later_fd(~ {}, fd1, timeout = 0)
  expect_false(loop_empty())
  run_now(1)
  expect_true(loop_empty())
})

local({
  # later_fd() errors when passed destroyed loops ----

  loop <- create_loop()
  destroy_loop(loop)
  expect_error(later_fd(identity, loop = loop))
})

#' @title Get and set logging level
#'
#' @description The logging level for later can be set to report differing
#' levels of information. Possible logging levels (from least to most
#' information reported) are: \code{"OFF"}, \code{"ERROR"}, \code{"WARN"},
#' \code{"INFO"}, or \code{"DEBUG"}. The default level is \code{ERROR}.
#'
#' @param level The logging level. Must be one of \code{NULL}, \code{"OFF"},
#'   \code{"ERROR"}, \code{"WARN"}, \code{"INFO"}, or \code{"DEBUG"}. If
#'   \code{NULL} (the default), then this function simply returns the current
#'   logging level.
#'
#' @return If \code{level=NULL}, then this returns the current logging level. If
#'   \code{level} is any other value, then this returns the previous logging
#'   level, from before it is set to the new value.
#'
#' @keywords internal
logLevel <- function(level = NULL) {
  if (is.null(level)) {
    level <- ""
    log_level("")
  } else {
    level <- match.arg(level, c("OFF", "ERROR", "WARN", "INFO", "DEBUG"))
    invisible(log_level(level))
  }
}

local({
  # logLevel works ----
  
  current <- later2:::logLevel()
  expect_true(current %in% c("OFF", "ERROR", "WARN", "INFO", "DEBUG"))

  previous <- later2:::logLevel("DEBUG")
  expect_equal(previous, current)
  expect_equal(later2:::logLevel(), "DEBUG")

  expect_equal(later2:::logLevel(current), "DEBUG")
  expect_equal(later2:::logLevel(), current)
})

# Simulate a user interrupt without depending on rlang. tryCatch()'s
# `interrupt =` handler is an exiting handler, so signalling this condition
# unwinds execution just like a real interrupt would.
simulate_interrupt <- function() {
    signalCondition(
        structure(
            class = c("interrupt", "condition"),
            list(message = "", call = sys.call(-1))
        )
    )
}

---
title: "Using later from C++"
author: "Mauricio Vargas Sepulveda"
date: "`r Sys.Date()`"
output:
  litedown::html_format:
    options:
      toc: true
      number_sections: true
vignette: >
  %\VignetteIndexEntry{02 - Setup}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
editor:
  markdown:
    wrap: sentence
---

# Using later from C++

You can call `later2::later` from C++ code in R packages. This is safe to call from either the main R thread or a different
thread, and for both cases the callback will be invoked from the main R thread.

To use the C++ interface, you will need to:

* Add `later` to the `DESCRIPTION` file, under both `LinkingTo` and `Imports`
* Make sure that the `NAMESPACE` file has an `import(later)` entry.
* Add `#include <later2_api.h>` to the top of each C++ file that uses the later API.

## Executing a C function later

The `later2::later` function is accessible from `later2_api.h` and its prototype looks like this:

```cpp
void later(void (*func)(void*), void* data, double secs)
```

The first argument is a pointer to a function that takes one `void*` argument and returns void. The second argument is a `void*`
that will be passed to the function when it is called back. And the third argument is the number of seconds to wait (at a minimum)
before invoking. In all cases, the function will be invoked on the R thread, when no user R code is executing.

## Background tasks

This package also offers a higher-level C++ helper class called `later2::BackgroundTask`, to make it easier to execute tasks on a
background thread. It takes care of launching the background thread for you, and returning control back to the R thread at a later
point. You are responsible for providing the actual code that executes on the background thread, as well as code that executes on
the R thread before and after the background task completes.

The public/protected interface looks like this:

```cpp
class BackgroundTask {

public:
  BackgroundTask();
  virtual ~BackgroundTask();

  // Start executing the task  
  void begin();

protected:
  // The task to be executed on the background thread.
  // Neither the R runtime nor any R data structures may be touched from the
  // background thread; any values that need to be passed into or out of the
  // Execute method must be included as fields on the Task subclass object.
  virtual void execute() = 0;
  
  // A short task that runs on the main R thread after the background task has
  // completed. It is safe to access the R runtime and R data structures from
  // here.
  virtual void complete() = 0;
}
```

Create your own subclass, implementing a custom constructor plus the `execute` and `complete` methods.

It is critical that the code in your `execute` method not mutate any R data structures, call any R code, or cause any R
allocations, as it will execute in a background thread where such operations are unsafe. You can, however, perform such operations
in the constructor (assuming you perform construction only from the main R thread) and `complete` method. Pass values between the
constructor and methods using fields.

```cpp
#include "cpp4r.hpp"
#include <later2_api.h>

using namespace cpp11;

class MyTask : public later2::BackgroundTask {
public:
  MyTask(doubles vec) :
    inputVals(as_cpp<std::vector<double> >(vec)) {
  }

protected:
  void execute() {
    double sum = 0;
    for (std::vector<double>::const_iterator it = inputVals.begin();
      it != inputVals.end();
      it++) {
      
      sum += *it;
    }
    result = sum / inputVals.size();
  }
  
  void complete() {
    Rprintf("Result is %f\n", result);
  }

private:
  std::vector<double> inputVals;
  double result;
};
```

To run the task, `new` up your subclass and call `begin()`, e.g. `(new MyTask(vec))->begin()`. There is no need to keep track of
the pointer, the task object will delete itself when the task is complete.

```r
[[cpp11::register]] void asyncMean(doubles data) {
  (new MyTask(data))->begin();
}
```

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{create_loop}
\alias{create_loop}
\alias{destroy_loop}
\alias{exists_loop}
\alias{current_loop}
\alias{with_temp_loop}
\alias{with_loop}
\alias{global_loop}
\title{Private event loops}
\usage{
create_loop(parent = current_loop())
destroy_loop(loop)
exists_loop(loop)
current_loop()
with_temp_loop(expr)
with_loop(loop, expr)
global_loop()
}

\arguments{
\item{loop}{A handle to an event loop.}

\item{expr}{An expression to evaluate.}

\item{parent}{The parent event loop for the one being created. Whenever the
  parent loop runs, this loop will also automatically run, without having to
  manually call \code{\link{run_now}()} on this loop. If \code{NULL}, then
  this loop will not have a parent event loop that automatically runs it; the
  only way to run this loop will be by calling \code{\link{run_now}()} on this
  loop.}
}

\value{
An object of class `event_loop`.
}

\description{
Normally, later uses a global event loop for scheduling and
running functions. However, in some cases, it is useful to create a
\emph{private} event loop to schedule and execute tasks without disturbing
the global event loop. For example, you might have asynchronous code that
queries a remote data source, but want to wait for a full back-and-forth
communication to complete before continuing in your code -- from the caller's
perspective, it should behave like synchronous code, and not do anything with
the global event loop (which could run code unrelated to your operation). To
do this, you would run your asynchronous code using a private event loop.

\code{create_loop} creates and returns a handle to a private event loop,
which is useful when for scheduling tasks when you do not want to interfere
with the global event loop.

\code{destroy_loop} destroys a private event loop.

\code{exists_loop} reports whether an event loop exists -- that is, that it
has not been destroyed.

\code{current_loop} returns the currently-active event loop. Any calls to
\code{\link{later}()} or \code{\link{run_now}()} will use the current loop by
default.

\code{with_loop} evaluates an expression with a given event loop as the
currently-active loop.

\code{with_temp_loop} creates an event loop, makes it the current loop, then
evaluates the given expression. Afterwards, the new event loop is destroyed.

\code{global_loop} returns a handle to the global event loop.
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{loop_empty}
\alias{loop_empty}
\title{Check if later loop is empty}
\usage{
loop_empty(loop = current_loop())
}

\arguments{
\item{loop}{A handle to an event loop.}
}

\value{
A logical value.
}

\description{
Returns true if there are currently no callbacks that are scheduled to
execute in the present or future.
}

\keyword{internal}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{later_fd}
\alias{later_fd}
\title{Executes a function when a file descriptor is ready}
\usage{
later_fd(
  func,
  readfds = integer(),
  writefds = integer(),
  exceptfds = integer(),
  timeout = Inf,
  loop = current_loop()
)
}

\arguments{
\item{func}{A function that takes a single argument, a logical vector that
  indicates which file descriptors are ready (a concatenation of `readfds`,
  `writefds` and `exceptfds`). This may be all `FALSE` if the
  `timeout` argument is non-`Inf`. File descriptors with error conditions
  pending are represented as `NA`, as are invalid file descriptors such as
  those already closed.}

\item{readfds}{Integer vector of file descriptors, or Windows SOCKETs, to
  monitor for being ready to read.}

\item{writefds}{Integer vector of file descriptors, or Windows SOCKETs, to
  monitor being ready to write.}

\item{exceptfds}{Integer vector of file descriptors, or Windows SOCKETs, to
  monitor for error conditions pending.}

\item{timeout}{Number of seconds to wait before giving up, and calling `func`
  with all `FALSE`. The default `Inf` implies waiting indefinitely.
  Specifying `0` will check once without blocking, and supplying a negative
  value defaults to a timeout of 1s.}

\item{loop}{A handle to an event loop. Defaults to the currently-active loop.}
}

\value{
A delayed function evaluation.
}

\description{
Schedule an R function or formula to run after an indeterminate
amount of time when file descriptors are ready for reading or writing,
subject to an optional timeout.

On the occasion the system-level `poll` (on Windows `WSAPoll`) returns an
error, the callback will be made on a vector of all `NA`s. This is
indistinguishable from a case where the `poll` succeeds but there are error
conditions pending against each file descriptor. If no file descriptors are
supplied, the callback is scheduled for immediate execution and made on the
empty logical vector `logical(0)`.
}

\examples{
\dontrun{
# Unix-only example

# Use the base R 'parallel' package to fork child processes and obtain
# real, pollable file descriptors. A child's `fd` becomes ready for
# reading once the child finishes running.

# 1. timeout: prints FALSE, FALSE
job1 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
job2 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
fd1 <- job1$fd[1]
fd2 <- job2$fd[1]
later_fd(print, c(fd1, fd2), timeout = 0.1)
Sys.sleep(0.2)
run_now()

# 2. fd1 ready: prints TRUE, FALSE
job1 <- parallel::mcparallel(TRUE)
fd1 <- job1$fd[1]
Sys.sleep(0.1)
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 3. both ready: prints TRUE, TRUE
job2 <- parallel::mcparallel(TRUE)
fd2 <- job2$fd[1]
Sys.sleep(0.1)
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 4. fd2 ready: prints FALSE, TRUE
parallel::mccollect(job1)
job1 <- parallel::mcparallel({
  Sys.sleep(1)
  TRUE
})
fd1 <- job1$fd[1]
later_fd(print, c(fd1, fd2), timeout = 1)
Sys.sleep(0.1)
run_now()

# 5. fds invalid: prints NA, NA
parallel::mccollect(job1)
parallel::mccollect(job2)
later_fd(print, c(fd1, fd2), timeout = 0)
Sys.sleep(0.1)
run_now()
}
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in promise.R
\name{promise}
\alias{promise}
\title{Create a new promise object}
\usage{
promise(action)
}

\arguments{
\item{action}{A function with signature `function(resolve, reject)`.}
}

\value{
A promise object (see [then()]).
}

\description{
`promise()` creates a new promise, synchronized with later2's
event loop (instead of the `later` package's event loop, as the `promises`
package does). A promise is a placeholder object for the eventual result (or
error) of an asynchronous operation.

The `action` function should be a piece of code that returns quickly, but
initiates a potentially long-running, asynchronous task. If/when the task
successfully completes, call `resolve(value)` where `value` is the result of
the computation. If the task fails, call `reject(reason)`, where `reason` is
either an error object or a character string.
}

\examples{
p1 <- promise(function(resolve, reject) {
  later(function() resolve(runif(1)), delay = 2)
})
then(p1, print)
run_now(3)
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in promise.R
\name{then}
\alias{then}
\alias{catch}
\alias{finally}
\title{Access the results of a promise}
\usage{
then(promise, onFulfilled = NULL, onRejected = NULL)
catch(promise, onRejected)
finally(promise, onFinally)
}

\arguments{
\item{promise}{A promise object.}

\item{onFulfilled}{A function to be invoked if `promise` resolves
  successfully. Called with the resolved value, and optionally a second
  `.visible` argument indicating whether the value is
  [visible][base::invisible()].}

\item{onRejected}{A function taking a single `reason` argument, to be
  invoked if `promise` fails.}

\item{onFinally}{A function with no arguments, called when `promise`
  either succeeds or fails.}
}

\value{
A new promise.
}

\description{
Use `then()` to access the eventual result of a promise (or, if
the operation fails, the reason for that failure). The call to `then()` is
non-blocking: it returns immediately, and the return value is itself a
new promise.

`catch()` is equivalent to `then()`, but without the `onFulfilled`
argument; it is typically used at the end of a promise chain to perform
error handling.

`finally()` is similar to `then()`, but takes a single no-argument
function that is executed upon completion of the promise, regardless of
whether the result is success or failure. The return value of the
`onFinally` callback is ignored; an error thrown from it is propagated
forward into the returned promise.
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{list_queue}
\alias{list_queue}
\title{Get the contents of an event loop, as a list}
\usage{
list_queue(loop = current_loop())
}

\description{
This function is for debugging only.
}

\keyword{internal}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in promise.R
\name{print.promise}
\alias{print.promise}
\title{Print a promise}
\usage{
\method{print}{promise}(x, ...)
}

\arguments{
\item{x}{A promise object.}

\item{...}{Further arguments passed to or from other methods (just for S3
complicance, unused)}
}

\value{
`NULL`, invisibly. Called for its side effect of printing.
}

\description{
Print a promise
}

\keyword{internal}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{next_op_secs}
\alias{next_op_secs}
\title{Relative time to next scheduled operation}
\usage{
next_op_secs(loop = current_loop())
}

\arguments{
\item{loop}{A handle to an event loop.}
}

\value{
A numeric value.
}

\description{
Returns the duration between now and the earliest operation that
is currently scheduled, in seconds. If the operation is in the past, the
value will be negative. If no operation is currently scheduled, the value
will be `Inf`.
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in promise.R
\name{is.promise}
\alias{is.promise}
\title{Determine whether an object is a promise}
\usage{
is.promise(x)
}

\arguments{
\item{x}{An R object to test.}
}

\value{
`TRUE` if `x` is a promise object created by [promise()], `FALSE`
  otherwise.
}

\description{
Determine whether an object is a promise
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{later}
\alias{later}
\title{Executes a function later}
\usage{
later(func, delay = 0, loop = current_loop())
}

\arguments{
\item{func}{A function or one-sided formula (e.g. `~ .x + 1`, using `.` or
  `.x` as the argument pronoun).}

\item{delay}{Number of seconds in the future to delay execution. There is no
  guarantee that the function will be executed at the desired time, but it
  should not execute earlier.}

\item{loop}{A handle to an event loop. Defaults to the currently-active loop.}
}

\value{
A function, which, if invoked, will cancel the callback. The
  function will return \code{TRUE} if the callback was successfully
  cancelled and \code{FALSE} if not (this occurs if the callback has
  executed or has been cancelled already).
}

\description{
Schedule an R function or formula to run after a specified
period of time. Similar to JavaScript's `setTimeout` function. Like
JavaScript, R is single-threaded so there's no guarantee that the operation
will run exactly at the requested time, only that at least that much time
will elapse.

The mechanism used by this package is inspired by Simon Urbanek's
[background](https://github.com/s-u/background) package and similar code in
Rhttpd.

To avoid bugs due to reentrancy, by default, scheduled operations only run
when there is no other R code present on the execution stack; i.e., when R is
sitting at the top-level prompt. You can force past-due operations to run at
a time of your choosing by calling [run_now()].

Error handling is not particularly well-defined and may change in the future.
options(error=browser) should work and errors in `func` should generally not
crash the R process, but not much else can be said about it at this point.
If you must have specific behavior occur in the face of errors, put error
handling logic inside of `func`.
}

\examples{
# Example of formula style
later(~ cat("Hello from the past\n"), 3)

# Example of function style
later(function() {
  print(summary(cars))
}, 2)
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in debug.R
\name{logLevel}
\alias{logLevel}
\title{Get and set logging level}
\usage{
logLevel(level = NULL)
}

\arguments{
\item{level}{The logging level. Must be one of \code{NULL}, \code{"OFF"},
  \code{"ERROR"}, \code{"WARN"}, \code{"INFO"}, or \code{"DEBUG"}. If
  \code{NULL} (the default), then this function simply returns the current
  logging level.}
}

\value{
If \code{level=NULL}, then this returns the current logging level. If
  \code{level} is any other value, then this returns the previous logging
  level, from before it is set to the new value.
}

\description{
The logging level for later can be set to report differing
levels of information. Possible logging levels (from least to most
information reported) are: \code{"OFF"}, \code{"ERROR"}, \code{"WARN"},
\code{"INFO"}, or \code{"DEBUG"}. The default level is \code{ERROR}.
}

\keyword{internal}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2-package.R
\name{later2-package}
\alias{later2-package}
\title{Utilities for Scheduling Functions to Execute Later with Event Loops}
\description{
Executes arbitrary R or C functions some time after the current
 time, after the R execution stack has emptied. The functions are scheduled
 in an event loop. This is a derived work from the 'later' package aiming
 to reduce the number of dependencies.
}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{print.event_loop}
\alias{print.event_loop}
\title{Print an event loop}
\usage{
\method{print}{event_loop}(x, ...)
}

\arguments{
\item{x}{An event loop object.}

\item{...}{Further arguments passed to or from other methods (currently
  unused).}
}

\value{
`NULL`, invisibly. Called for its side effect of printing.
}

\description{
Print an event loop
}

\keyword{internal}

% Generated by tinyroxygen: do not edit by hand
% Please edit documentation in later2.R
\name{run_now}
\alias{run_now}
\title{Execute scheduled operations}
\usage{
run_now(timeoutSecs = 0L, all = TRUE, loop = current_loop())
}

\arguments{
\item{timeoutSecs}{Wait (block) for up to this number of seconds waiting for
  an operation to be ready to run. If `0`, then return immediately if there
  are no operations that are ready to run. If `Inf` or negative, then wait as
  long as it takes (if none are scheduled, then this will block forever).}

\item{all}{If `FALSE`, `run_now()` will execute at most one scheduled
  operation (instead of all eligible operations). This can be useful in cases
  where you want to interleave scheduled operations with your own logic.}

\item{loop}{A handle to an event loop. Defaults to the currently-active loop.}
}

\value{
A logical indicating whether any callbacks were actually run.
}

\description{
Normally, operations scheduled with [later()] will not execute
unless/until no other R code is on the stack (i.e. at the top-level). If you
need to run blocking R code for a long time and want to allow scheduled
operations to run at well-defined points of your own operation, you can call
`run_now()` at those points and any operations that are due to run will do
so.

If one of the callbacks throws an error, the error will _not_ be caught, and
subsequent callbacks will not be executed (until `run_now()` is called again,
or control returns to the R prompt). You must use your own
[tryCatch][base::conditions] if you want to handle errors.
}
